@@ -37,10 +37,10 @@ class TestAutovivification < LoaderTest
 
   test "autoloads several constants from the same namespace (Object)" do
     files = [
-      ["app/models/admin/hotel.rb", "class Admin::Hotel; end"],
-      ["app/controllers/admin/hotels_controller.rb", "class Admin::HotelsController; end"]
+      ["rd1/admin/hotel.rb", "class Admin::Hotel; end"],
+      ["rd2/admin/hotels_controller.rb", "class Admin::HotelsController; end"]
     ]
-    with_setup(files, dirs: %w(app/models app/controllers)) do
+    with_setup(files) do
       assert Admin::Hotel
       assert Admin::HotelsController
     end
@@ -48,16 +48,27 @@ class TestAutovivification < LoaderTest
 
   test "autoloads several constants from the same namespace (Namespace)" do
     files = [
-      ["app/models/admin/hotel.rb", "class #{Namespace}::Admin::Hotel; end"],
-      ["app/controllers/admin/hotels_controller.rb", "class #{Namespace}::Admin::HotelsController; end"]
+      ["rd1/admin/hotel.rb", "class #{Namespace}::Admin::Hotel; end"],
+      ["rd2/admin/hotels_controller.rb", "class #{Namespace}::Admin::HotelsController; end"]
     ]
-    with_setup(files, namespace: Namespace, dirs: %w(app/models app/controllers)) do
+    with_setup(files, namespace: Namespace) do
       assert Namespace::Admin::Hotel
       assert Namespace::Admin::HotelsController
     end
   end
 
+  test "does not register the namespace as explicit" do
+    files = [
+      ["rd1/admin/x.rb", "Admin::X = true"],
+      ["rd2/admin/y.rb", "Admin::Y = true"]
+    ]
+    with_setup(files) do
+      assert !Zeitwerk::ExplicitNamespace.__registered?("Admin")
+    end
+  end
+
   test "autovivification is synchronized" do
+    $test_admin_const_set_calls = 0
     $test_admin_const_set_queue = Queue.new
 
     files = [["admin/v2/user.rb", "class Admin::V2::User; end"]]
@@ -65,8 +76,9 @@ class TestAutovivification < LoaderTest
       assert Admin
 
       def Admin.const_set(cname, mod)
+        $test_admin_const_set_calls += 1
         $test_admin_const_set_queue << true
-        sleep 0.5
+        sleep 0.1
         super
       end
 
@@ -83,6 +95,7 @@ class TestAutovivification < LoaderTest
       concurrent_autovivifications.each(&:join)
 
       assert $test_admin_const_set_queue.empty?
+      assert_equal 1, $test_admin_const_set_calls
     end
   end
 
@@ -119,19 +132,13 @@ class TestAutovivification < LoaderTest
   end
 
   test "defines no namespace for directories whose Ruby files are all ignored" do
-    with_files([["foo/bar/baz.rb", ""]]) do
-      loader.push_dir(".")
-      loader.ignore("foo/bar/baz.rb")
-      loader.setup
+    with_setup([["foo/bar/ignored.rb", "IGNORED"]]) do
       assert !Object.autoload?(:Foo)
     end
   end
 
   test "defines no namespace for directories that have Ruby files below ignored directories" do
-    with_files([["foo/bar/baz.rb", ""]]) do
-      loader.push_dir(".")
-      loader.ignore("foo/bar")
-      loader.setup
+    with_setup([["foo/ignored/baz.rb", "IGNORED"]]) do
       assert !Object.autoload?(:Foo)
     end
   end
