@@ -2,6 +2,6 @@
 
 module RemoveConst
   def remove_const(cname, from: Object)
-    from.send(:remove_const, cname)
+    from.__send__(:remove_const, cname)
   end
 end
